//
// Copyright (c) 2023 Fas Xmut (fasxmut at protonmail dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

// BNirtDeviceSFML.hpp (20230731)

#ifndef __src_nirtcpp_BNirtDeviceSFML_hpp__
#define __src_nirtcpp_BNirtDeviceSFML_hpp__

#include <nirtcpp/core/engine/NirtCompileConfig.hpp>

#ifdef _NIRT_COMPILE_WITH_SFML_DEVICE_

#include <nirtcpp/core/engine/NirtcppDevice.hpp>
#include "CIrrDeviceStub.hpp"
#include "IImagePresenter.hpp"
#include "os.hpp"
#include <nirtcpp/core/engine/ICursorControl.hpp>

// sfml-todo: None define conflicts with x11
#ifdef None
#undef None
#endif
#include <SFML/Graphics.hpp>
#include <SFML/System.hpp>
#include <map>
#include <memory>

/*
SFML is controlled by macro:
_NIRT_COMPILE_WITH_SFML_DEVICE_

Unfinished Files (in src/nirtcpp):
	COpenGLCommon.hpp
	COpenGLDriver.cpp // Implement sfml-gl-swapbuffers
	os.cpp // Implement bswap_16, bswap_32 macro
	BNirtDeviceSFML.hpp
	BNirtDeviceSFML.cpp

*/

namespace nirt
{

class BNirtDeviceSFML: public nirt::CIrrDeviceStub, public nirt::video::IImagePresenter
{
private:
	std::unique_ptr<sf::RenderWindow> sfml_window;
	sf::WindowHandle sfml_screen; // sfml-todo
	std::unique_ptr<nirt::u8 *> SoftwareImage; // For EDT_SOFTWARE and EDT_BURNINGSVIDEO // sfml-todo: need implementation
private:
	nirt::s32 MouseX, MouseY;
	nirt::u32 MouseButtonStates;
	nirt::u32 Width, Height;
	bool Resizable;
	bool WindowMinimized;
private:
	// SKeyMap: first: nirtcpp key codes; second: sfml key codes.
	// Key Codes are defined in nirtcpp/Keycodes.hpp
	using SKeyMap = std::map<nirt::s32, nirt::EKEY_CODE>; // sfml-todo: use std::map
	SKeyMap KeyMap;
private:
	// sfml_sys_wm_info info; // sfml-todo

private:
	void createDriver();
	// Key Codes are defined in nirtcpp/Keycodes.hpp
	void createKeyMap();

private:
// create window
		// create window by different driver type.
	bool createWindow();
		// If the driver is EDT_OPENGL, createWindow will call this method.
	bool createSFMLGLWindow();
		// short hand of sfml_window->create(???)
		// This method only do sfml_window->create, and does not do any checks.
	void create_sfml_gl_window_proxy();
		// If create_sfml_gl_window_proxy() is called, and sfml_window->isOpen() returns true, call this method.
		// It does not do any checks.
	void on_create_sfml_gl_window();

private:
		// enum =>
		//			sf::Style::None=0, sf::Style::Titlebar, sf::Style::Resize,
		//			sf::Style::Close, sf::Style::Fullscreen, sf::Style::Default
		//		sf::Style::Fullscreen flag and all others are mutually exclusive.
		//		sf::Style::None flag and all others are mutually exclusive.
	nirt::u32 sfml_style; // It is generated by information from this->CreationParams

		// Generate sfml_style by information from this->CreationParams.
	void init_sfml_style();

		// Generate sfml_style by its params, it does not change information of this->CreationParams.
	void update_sfml_style(bool new_resizable);

private:
	bool sfml_window_focused = false;

private:
#if defined(_NIRT_COMPILE_WITH_JOYSTICK_EVENTS_)
	//nirt::core::array<sf::Joystick *> Joysticks; // sfml-todo
#endif

public:
	// Implement cursor control with SFML
	class CCursorControl: public nirt::gui::ICursorControl
	{
	private:
		BNirtDeviceSFML * Device;
		bool Null;
		bool IsVisible;
		nirt::core::position2d<nirt::s32> CursorPos;
		nirt::core::rect<nirt::s32> ReferenceRect;
		bool UseReferenceRect;
		bool LastQuery;
	private:
		inline void updateCursorPos()
		{
			if (this->Null)
				return;
			if (! os::Timer::isStopped())
			{
				nirt::u32 now = os::Timer::getTime();
				if (now <= LastQuery)
					return;
				LastQuery = now;
			}
			CursorPos.X = Device->MouseX;
			CursorPos.Y = Device->MouseY;
		}
	public:
	// constructors
		CCursorControl(BNirtDeviceSFML * dev, bool null):
			Device{dev},
			Null{null},
			IsVisible{true},
			UseReferenceRect{false},
			LastQuery{0}
		{
		}
	public:
	// mouse cursor visible
		// Change mouse cursor visible state
		inline void setVisible(bool visible) override
		{
			if (visible == this->IsVisible)
				return;
			this->IsVisible = visible;
			if (! this->Null)
			{
				this->Device->sfml_window->setMouseCursorVisible(visible); // sfml-todo
				this->Device->sfml_window->setMouseCursorGrabbed(! visible);
			}
		}
		bool isVisible() const override
		{
			return IsVisible;
		}
	public:
	// set mouse cursor position
		inline void setPosition(const nirt::core::position2d<nirt::f32> & pos) override
		{
			this->setPosition(pos.X, pos.Y);
		}
		inline void setPosition(nirt::f32 x, nirt::f32 y) override
		{
			this->setPosition(
				static_cast<nirt::s32>(x*Device->Width),
				static_cast<nirt::s32>(y*Device->Height)
			);
		}
		inline void setPosition(const nirt::core::position2d<nirt::s32> & pos) override
		{
			this->setPosition(pos.X, pos.Y);
		}
		inline void setPosition(nirt::s32 x, nirt::s32 y) override
		{
			if (! this->Null)
			{
				if (UseReferenceRect)
				{
					sf::Mouse::setPosition(
						sf::Vector2i{
							ReferenceRect.UpperLeftCorner.X + x,
							ReferenceRect.UpperLeftCorner.Y + y
						},
						* Device->sfml_window
					);
				}
				else
				{
					sf::Mouse::setPosition(
						sf::Vector2i{
							x,
							y
						},
						* Device->sfml_window
					);
				}
			} // if (! this->Null)

			CursorPos.X = x;
			CursorPos.Y = y;
		}
	public:
	// Get mouse cursor position
		inline const nirt::core::position2d<nirt::s32> & getPosition(bool updateCursor) override
		{
			if (updateCursor)
				this->updateCursorPos();
			return this->CursorPos;
		}
		inline nirt::core::position2d<nirt::f32> getRelativePosition(bool updateCursor) override
		{
			if (updateCursor)
				this->updateCursorPos();
			if (! UseReferenceRect)
			{
				return nirt::core::position2d<nirt::f32>{
					CursorPos.X / static_cast<nirt::f32>(Device->Width),
					CursorPos.Y / static_cast<nirt::f32>(Device->Height)
				};
			}

			// else
			return nirt::core::position2d<nirt::f32>{
				CursorPos.X / static_cast<nirt::f32>(ReferenceRect.getWidth()),
				CursorPos.Y / static_cast<nirt::f32>(ReferenceRect.getHeight())
			};
		}
	public:
		inline void setReferenceRect(nirt::core::rect<nirt::s32> * rect = nullptr) override
		{
			// sfml-todo
			if (rect)
			{
				ReferenceRect = *rect;
				UseReferenceRect = true;
				if (! ReferenceRect.getHeight() || ReferenceRect.getHeight()%2)
					ReferenceRect.LowerRightCorner.Y += 1;
				if (! ReferenceRect.getWidth() || ReferenceRect.getWidth()%2)
					ReferenceRect.LowerRightCorner.X += 1;
			}
			else
			{
				UseReferenceRect = false;
			}
		}
	}; // class CCursorControl

public:
	virtual ~BNirtDeviceSFML();
	BNirtDeviceSFML(const nirt::SNirtcppCreationParameters & param);

public:
	bool run() override;
	void yield() override;
	void sleep(nirt::u32 time_ms, bool pause_timer=false) override;
	void setWindowCaption(const wchar_t * text) override;
	bool isWindowActive() const override;
	bool isWindowFocused() const override;
	bool isWindowMinimized() const override;
	nirt::video::ECOLOR_FORMAT getColorFormat() const override;
	bool present(nirt::video::IImage * src_surface, void * dst_window_id=0, nirt::core::rect<s32> * src_clip=0) override;
	void closeDevice() override;
	nirt::video::IVideoModeList* getVideoModeList() override;
		//void setResizable(bool new_resizable=false) override;
	// Limits: setResizable(?)  must be called before drawing, adding scene/gui objects.
	void setResizable(bool new_resizable) override; // sfml-todo: no default value, need test
	void minimizeWindow() override;
	void maximizeWindow() override;
	void restoreWindow() override;
	nirt::core::position2di getWindowPosition() override;
	bool activateJoysticks(core::array<nirt::SJoystickInfo> & joystick_info) override;
	bool setGammaRamp(nirt::f32 red, nirt::f32 green, nirt::f32 blue, nirt::f32 brightness, nirt::f32 contrast) override;
	bool getGammaRamp(nirt::f32 &red, nirt::f32 &green, nirt::f32 & blue, nirt::f32 & brightness, nirt::f32 & contrast) override;
	// sfml-finished
	inline nirt::E_DEVICE_TYPE getType() const override
	{
		return nirt::EIDT_SFML;
	}
	// This is identically calling sfml_window->display()
	void sfml_display();
}; // class BNirtDeviceSFML

} // namespace nirt

#endif // _NIRT_COMPILE_WITH_SFML_DEVICE_

#endif // __src_nirtcpp_BNirtDeviceSFML_hpp__

