// Copyright (C) 2002-2012 Nikolaus Gebhardt
// This file is part of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in nirtcpp/nirtcpp.hpp

#ifndef NIRT_C_PAK_READER_HPP_INCLUDED
#define NIRT_C_PAK_READER_HPP_INCLUDED

#include <nirtcpp/core/engine/NirtCompileConfig.hpp>

#ifdef __NIRT_COMPILE_WITH_PAK_ARCHIVE_LOADER_

#include <nirtcpp/core/engine/IReferenceCounted.hpp>
#include <nirtcpp/core/engine/IReadFile.hpp>
#include <nirtcpp/core/engine/irrArray.hpp>
#include <nirtcpp/core/engine/irrString.hpp>
#include <nirtcpp/core/engine/IFileSystem.hpp>
#include "CFileList.hpp"

namespace nirt
{
namespace io
{
	//! File header containing location and size of the table of contents
	class SPAKFileHeader
	{
	public:
		// Don't change the order of these fields!  They must match the order stored on disk.
		c8 tag[4];
		u32 offset;
		u32 length;
	};

	//! An entry in the PAK file's table of contents.
	class SPAKFileEntry
	{
	public:
		// Don't change the order of these fields!  They must match the order stored on disk.
		c8 name[56];
		u32 offset;
		u32 length;
	};

	//! Archiveloader capable of loading PAK Archives
	class CArchiveLoaderPAK : public IArchiveLoader
	{
	public:

		//! Constructor
		CArchiveLoaderPAK(io::IFileSystem* fs);

		//! returns true if the file maybe is able to be loaded by this class
		//! based on the file extension (e.g. ".zip")
		virtual bool isALoadableFileFormat(const io::path& filename) const override;

		//! Check if the file might be loaded by this class
		/** Check might look into the file.
		\param file File handle to check.
		\return True if file seems to be loadable. */
		virtual bool isALoadableFileFormat(io::IReadFile* file) const override;

		//! Check to see if the loader can create archives of this type.
		/** Check based on the archive type.
		\param fileType The archive type to check.
		\return True if the archile loader supports this type, false if not */
		virtual bool isALoadableFileFormat(E_FILE_ARCHIVE_TYPE fileType) const override;

		//! Creates an archive from the filename
		/** \param file File handle to check.
		\return Pointer to newly created archive, or 0 upon error. */
		virtual IFileArchive* createArchive(const io::path& filename, bool ignoreCase, bool ignorePaths) const override;

		//! creates/loads an archive from the file.
		//! \return Pointer to the created archive. Returns 0 if loading failed.
		virtual io::IFileArchive* createArchive(io::IReadFile* file, bool ignoreCase, bool ignorePaths) const override;

	private:
		io::IFileSystem* FileSystem;
	};


	//! reads from pak
	class CPakReader : public virtual IFileArchive, virtual CFileList
	{
	public:

		CPakReader(IReadFile* file, bool ignoreCase, bool ignorePaths);
		virtual ~CPakReader();

		// file archive methods

		//! return the id of the file Archive
		virtual const io::path& getArchiveName() const override
		{
			return File->getFileName();
		}

		//! opens a file by file name
		virtual IReadFile* createAndOpenFile(const io::path& filename) override;

		//! opens a file by index
		virtual IReadFile* createAndOpenFile(u32 index) override;

		//! returns the list of files
		virtual const IFileList* getFileList() const override;

		//! get the class Type
		virtual E_FILE_ARCHIVE_TYPE getType() const override { return EFAT_PAK; }

	private:

		//! scans for a local header, returns false if the header is invalid
		bool scanLocalHeader();

		IReadFile* File;

	};

} // end namespace io
} // end namespace nirt

#endif // __NIRT_COMPILE_WITH_PAK_ARCHIVE_LOADER_

#endif // NIRT_C_PAK_READER_HPP_INCLUDED
